# Copyright 2022-2023 c0fec0de
#
# This file is part of Git Workspace.
#
# Git Workspace is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Git Workspace is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Git Workspace. If not, see <https://www.gnu.org/licenses/>.

"""
Google's git-repo replacement powered by git-ws.
"""

from .repomanifestformat import RepoManifestFormat



# Copyright 2022-2023 c0fec0de
#
# This file is part of Git Workspace.
#
# Git Workspace is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Git Workspace is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Git Workspace. If not, see <https://www.gnu.org/licenses/>.

"""Logging."""
import logging

import click

_LOGLEVELMAP = {
    0: logging.WARNING,
    1: logging.INFO,
    2: logging.DEBUG,
}


class LogHandler(logging.Handler):
    """Logging Handler."""

    def __init__(self):
        super().__init__()
        self.has_errors = False

    def emit(self, record):
        """Emit Message."""
        if record.levelno >= logging.ERROR:
            self.has_errors = True
        click.echo(self.format(record), err=record.levelno >= logging.WARNING)


class LogFormatter(logging.Formatter):
    """Log Formatter."""

    levelsettings = {  # noqa: RUF012
        logging.WARNING: {"fg": "yellow", "bold": True},
        logging.ERROR: {"fg": "red", "bold": True},
    }
    leveldefault = {"bold": True, "dim": True}  # noqa: RUF012
    msgsettings = {  # noqa: RUF012
        logging.DEBUG: {"dim": True, "fg": "cyan"},
        logging.INFO: {"dim": True, "fg": "blue"},
        logging.WARNING: {"fg": "yellow"},
        logging.ERROR: {"fg": "red"},
    }
    msgdefault = {"fg": "blue", "dim": True}  # noqa: RUF012

    def __init__(self, color: bool):
        super().__init__()
        self.color: bool = color

    def format(self, record):
        """Format Message."""
        msg = record.getMessage()
        if self.color:
            levelsetting = self.levelsettings.get(record.levelno, self.leveldefault)
            msgsetting = self.msgsettings.get(record.levelno, self.msgdefault)
            prefix = click.style(f"{record.levelname}:".ljust(8), **levelsetting)
            lines = [click.style(line, **msgsetting) for line in msg.splitlines()]
        else:
            prefix = f"{record.levelname}:".ljust(8)
            lines = msg.splitlines()

        return "\n".join(f"{prefix} {line}" for line in lines)


def setup_logging(color: bool, verbose: int) -> LogHandler:
    """Create And Setup Logging Handler."""
    handler = LogHandler()
    handler.formatter = LogFormatter(color)
    logger = logging.getLogger()
    level = _LOGLEVELMAP.get(verbose, logging.DEBUG)
    logger.setLevel(level)
    logger.handlers = [handler]
    return handler



# Copyright 2022-2023 c0fec0de
#
# This file is part of Git Workspace.
#
# Git Workspace is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Git Workspace is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Git Workspace. If not, see <https://www.gnu.org/licenses/>.

"""Command Line Interface."""

import click
from gitws import AppConfig, AppConfigLocation, Defaults, GitWS, filter_clone_on_branch

# from gitws._manifestformatmanager import get_manifest_format_manager
# from gitws._util import resolve_relative
# from gitws.git import FileStatus, State
from .common import COLOR_INFO, Context, Error, exceptionhandling, pass_context
from .logging import setup_logging


def _version_option():  # pragma: no cover
    # Add support for click 7.x.x and click 8.x.x
    if click.version_option.__kwdefaults__ and "package_name" in click.version_option.__kwdefaults__:
        return click.version_option(package_name="git-ws")
    return click.version_option()


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.option("-v", "--verbose", count=True)
@_version_option()
@click.pass_context
def main(ctx=None, verbose=0):
    """
    Google's git-repo replacement powered by git-ws.
    """
    app_config = AppConfig()
    color = Error.color = app_config.options.color_ui
    handler = setup_logging(color, verbose)
    ctx.obj = Context(verbose=verbose, color=color, handler=handler)


@main.command()
@pass_context
def init(
    context,
):
    """
    Init.
    """
    with exceptionhandling(context):
        click.secho("Init")


@main.command()
@pass_context
def sync(
    context,
):
    """
    Sync.
    """
    with exceptionhandling(context):
        click.secho("Sync")



# Copyright 2022-2023 c0fec0de
#
# This file is part of Git Workspace.
#
# Git Workspace is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Git Workspace is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Git Workspace. If not, see <https://www.gnu.org/licenses/>.

"""Command Line Interface Utilities."""
import traceback
from contextlib import contextmanager
from typing import Any

import click
from pydantic import BaseModel

COLOR_INFO = "blue"


class Context(BaseModel):
    """Command Line Context."""

    verbose: int
    color: bool
    handler: Any = None

    def secho(self, message, **kwargs):
        """Print with color support similar to :any:`click.secho()."""
        if self.color:
            return click.secho(message, **kwargs)
        return click.echo(message)

    def style(self, text, **kwargs):
        """Format ``text``."""
        if self.color:
            return click.style(text, **kwargs)
        return text


pass_context = click.make_pass_decorator(Context)


class Error(click.ClickException):
    """Common CLI Error."""

    color = True

    def format_message(self) -> str:
        if self.color:
            return click.style(self.message, fg="red")
        return self.message


@contextmanager
def exceptionhandling(context: Context):
    """
    Click Exception Handling.

    The GitWS implementation shall NOT depend on click (except the cli module).
    Therefore we remap any internal errors to nice click errors.

    We provide useful command line hints, depending on exceptions.
    """
    try:
        yield
    except Exception as exc:
        _print_traceback(context)
        raise Error(f"{exc!s}") from None
    if context.handler.has_errors:
        context.secho("Aborted!", bold=True, fg="red")
        raise click.exceptions.Exit(1)


def _print_traceback(context: Context):
    if context.verbose > 1:  # pragma: no cover
        lines = "".join(traceback.format_exc())
        context.secho(lines, fg="red", err=True)



# Copyright 2022-2023 c0fec0de
#
# This file is part of Git Workspace.
#
# Git Workspace is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.
#
# Git Workspace is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with Git Workspace. If not, see <https://www.gnu.org/licenses/>.

"""
Google Git Repo Manifest Format.
"""
import re
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List
from xml.etree.ElementTree import tostring

from defusedxml import ElementTree
from gitws import (
    Defaults,
    FileRef,
    Group,
    ManifestError,
    ManifestFormat,
    ManifestNotFoundError,
    ManifestSpec,
    ProjectSpec,
    Remote,
    ValidationError,
)
from gitws._util import LOGGER

_RE_SPLIT = re.compile(r"[,\s]\s*")


class RepoManifestFormat(ManifestFormat):
    """
    Google Repo Manifest Format.
    """

    def is_compatible(self, path: Path) -> bool:
        """Check If v  File At ``path`` Is Compatible."""
        return path.suffix == ".xml"

    def load(self, path: Path) -> ManifestSpec:  # noqa: C901, PLR0915
        """
        Load Manifest From ``path``.

        Raises:
            ManifestNotFoundError: if file is not found
            ManifestError: On Syntax Or Data Scheme Errors.
        """
        defaults: Dict[str, str] = {}
        remotes: List[Remote] = []
        projects: List[ProjectSpec] = []
        ignored: List[str] = []

        # Feel free to re-factor

        def _convert_default(defaults: Dict[str, str], element):
            for name, value in element.attrib.items():
                if name in ("remote", "revision"):
                    defaults[name] = value
                else:
                    _ignore(f"default.{name}")

        def _convert_remote(remotes: List[Remote], element):
            remote = {}
            for name, value in element.attrib.items():
                if name == "name":
                    remote[name] = value
                elif name == "fetch":
                    remote["url-base"] = value
                else:
                    _ignore(f"remote.{name}")
            with _handle_validation_error(path, element):
                remotes.append(Remote(**remote))

        def _convert_project(projects: List[ProjectSpec], element, pname=None, ppath=None):
            copyfiles: List[FileRef] = []
            linkfiles: List[FileRef] = []
            groups: List[Group] = []
            project = {
                "recursive": False,
            }
            subprojects: List[ProjectSpec] = []
            for name, value in element.attrib.items():
                if name == "name":
                    project[name] = f"{pname}{value}" if pname else value
                elif name == "path":
                    project[name] = f"{ppath}/{value}" if ppath else value
                elif name in ("remote", "revision"):
                    project[name] = value
                elif name == "groups":
                    groups.extend(item.strip() for item in _RE_SPLIT.split(value))
                else:
                    _ignore(f"default.{name}")
            # group compatibility
            pname = project["name"]
            ppath = project.get("path", pname)
            # subelements
            for subelement in element:
                if subelement.tag == "copyfile":
                    _convert_file(copyfiles, "project.copyfile", subelement)
                elif subelement.tag == "linkfile":
                    _convert_file(linkfiles, "project.linkfile", subelement)
                elif subelement.tag == "project":
                    _convert_project(subprojects, subelement, pname=pname, ppath=ppath)
                else:
                    _ignore(f"project.{subelement.tag}")
            with _handle_validation_error(path, element):
                projects.append(
                    ProjectSpec(copyfiles=tuple(copyfiles), linkfiles=tuple(linkfiles), groups=tuple(groups), **project)
                )
            projects.extend(subprojects)

        def _convert_file(files: List[FileRef], prefix: str, element):
            file = {}
            for name, value in element.attrib.items():
                if name in ("src", "dest"):
                    file[name] = value
                else:
                    _ignore(f"{prefix}.{name}")
            with _handle_validation_error(path, element):
                files.append(FileRef(**file))

        def _ignore(name):
            if name not in ignored:
                ignored.append(name)
                LOGGER.info("%r: Ignoring %r", str(path), name)

        try:
            root = ElementTree.parse(path).getroot()
        except FileNotFoundError:
            raise ManifestNotFoundError(path) from None
        except Exception as exc:
            raise ManifestError(path, str(exc)) from None

        if root.tag != "manifest":
            raise ManifestError(path, f"Root element is {root.tag!r}. Expecting 'manifest'")

        for element in root:
            tag = element.tag
            if tag == "default":
                _convert_default(defaults, element)
            elif tag == "remote":
                _convert_remote(remotes, element)
            elif tag == "project":
                _convert_project(projects, element)
            else:
                _ignore(tag)

        with _handle_validation_error(path, root):
            return ManifestSpec(
                defaults=Defaults(**defaults),
                remotes=tuple(remotes),
                dependencies=tuple(projects),
                group_filters=["-notdefault"],
            )


@contextmanager
def _handle_validation_error(path, element):
    try:
        yield
    except ValidationError as exc:
        LOGGER.debug(str(exc))
        expr = tostring(element).decode("utf-8").strip()
        raise ManifestError(path, expr) from None
